#define foreach(C) for ((C).First(); (C).HasNext(); (C).Next())

namespace Obelisk
{
	typedef unsigned int TIntPtr;



	template <typename T> class CCollectionIndex
	{
	public:
		struct SEntry
		{
			SEntry()
					: Object(NULL), Next(NULL) { }
			SEntry(T* aObject)
					: Object(aObject), Next(NULL) { }

			T* Object;
			SEntry *Next;
		};



		CCollectionIndex()
		{
			mReferenceCount = 1;
			mCapacity = 8;
			mRehashCapacity = mCapacity * 0.8;
			mData = new SEntry [mCapacity];
			mNumItems = 0;
		}



		CCollectionIndex(CCollectionIndex<T>* aCollection)
		{
			mReferenceCount = 1;

			mCapacity = Max(aCollection->mNumItems * 2, 8u);
			mRehashCapacity = mCapacity * 0.8;
			mData = new SEntry [mCapacity];
			mNumItems = aCollection->mNumItems;

			for (unsigned int i = 0; i < aCollection->mCapacity; i++)
			{
				if (aCollection->mData[i].Object != NULL)
				{
					for (SEntry *Entry = &aCollection->mData[i]; Entry != NULL; Entry = Entry->Next)
						AddObject(Entry->Object);
				}
			}
		}



		~CCollectionIndex()
		{
			for (unsigned int i = 0; i < mCapacity; i++)
			{
				SEntry *Entry = mData[i].Next;
				while (Entry != NULL)
				{
					SEntry *Next = Entry->Next;
					delete Entry;
					Entry = Next;
				}
			}
			delete [] mData;
		}



		CCollectionIndex<T>* AddRef()
		{
			mReferenceCount++;
			return this;
		}



		void Release()
		{
			Assert(mReferenceCount >= 0)
			if (--mReferenceCount == 0)
				delete this;
		}



		int GetCount()
		{
			return mNumItems;
		}



		void Add(T* aObject)
		{
			AddObject(aObject);
			mNumItems++;
			if (mNumItems == mRehashCapacity)
				Rehash(mCapacity * 2);
		}



		void Remove(T* aObject)
		{
			Assert(aObject != NULL)
			unsigned int Index = ((TIntPtr)aObject) % mCapacity;
			SEntry *Last = NULL;
			for (SEntry* e = &mData[Index]; e != NULL; e = e->Next)
			{
				if (e->Object == aObject)
				{
					if (Last == NULL && e->Next == NULL)
						mData[Index] = SEntry();
					else if (Last == NULL)
					{
						SEntry *Next = e->Next;
						mData[Index] = *Next;
						delete Next;
					}
					else
					{
						Last->Next = e->Next;
						delete e;
					}
					mNumItems--;
					// TODO: rehash if NumEntries is below a threshold
					return;
				}
				Last = e;
			}
			throw CException("Object is not in collection");
		}



		bool IsShared()
		{
			return mReferenceCount != 1;
		}



		int GetCapacity()
		{
			return mCapacity;
		}



		void IteratorFirst(SEntry*& aEntry, unsigned int& aIndex)
		{
			for (unsigned int i = 0; i < mCapacity; i++)
			{
				for (SEntry* e = &mData[i]; e != NULL; e = e->Next)
				{
					if (e->Object != NULL)
					{
						aEntry = e;
						aIndex = i;
						return;
					}
				}
			}
			aEntry = NULL;
		}



		void IteratorNext(SEntry*& aEntry, SEntry*& aPrevEntry, unsigned int& aIndex)
		{
			if (aEntry == NULL)
			{
				if (aPrevEntry == NULL)
					aEntry = &mData[aIndex];
				else
					aEntry = aPrevEntry;
			}
			if (aEntry->Next != NULL)
			{
				aPrevEntry = aEntry;
				aEntry = aEntry->Next;
				return;
			}
			for (unsigned int i = aIndex + 1; i < mCapacity; i++)
			{
				for (SEntry* e = &mData[i]; e != NULL; e = e->Next)
				{
					if (e->Object != NULL)
					{
						aEntry = e;
						aPrevEntry = NULL;
						aIndex = i;
						return;
					}
				}
			}
			aEntry = NULL;
		}



		void IteratorDelete(SEntry* aEntry, SEntry* aPrevEntry, unsigned int aIndex)
		{
			if (aPrevEntry == NULL && aEntry->Next == NULL)
				mData[aIndex] = SEntry();
			else if (aPrevEntry == NULL)
			{
				SEntry *Next = aEntry->Next;
				mData[aIndex] = *Next;
				delete Next;
			}
			else
			{
				aPrevEntry->Next = aEntry->Next;
				delete aEntry;
			}
			mNumItems--;
		}



	private:
		int mReferenceCount;
		unsigned int mCapacity;
		unsigned int mRehashCapacity;
		SEntry *mData;
		unsigned int mNumItems;



		/*
		Increases the capacity of the table and rebuilds it.
		*/
		void Rehash(int aNewCapacity)
		{
			SEntry *OldData = mData;
			unsigned int OldCapacity = mCapacity;
			mCapacity = aNewCapacity;
			mData = new SEntry [mCapacity];

			for (unsigned int i = 0; i < OldCapacity; i++)
			{
				SEntry *e = &OldData[i];
				if (e->Object == NULL)
				{
					Assert(e->Next == NULL)
					continue;
				}
				bool First = true;
				do
				{
					SEntry *Next = e->Next;
					unsigned int Index = ((TIntPtr)e->Object) % mCapacity;
					SEntry *Last = &mData[Index];
					if (Last->Object == NULL)
					{
						Assert(mData[Index].Next == NULL)
						mData[Index] = SEntry(e->Object);
						if (!First)
							delete e;
					}
					else
					{
						while (Last->Next != NULL)
						{
							Assert(Last->Object != NULL)
							Last = Last->Next;
						}
						if (First)
							Last->Next = new SEntry(e->Object);
						else
						{
							e->Next = NULL;
							Last->Next = e;
						}
					}

					e = Next;
					Assert(e == NULL || e->Object != NULL)
					First = false;
				} while (e != NULL);
			}
			delete [] OldData;

			mRehashCapacity = mCapacity * 0.8;
		}



		/*
		bool ContainsEntry(T* aObject) const
		{
			Assert(aObject != NULL)
			unsigned int Index = ((TIntPtr)aObject) % mCapacity;
			for (SEntry* Entry = &mData[Index]; Entry != NULL; Entry = Entry->Next)
			{
				Assert((Entry->Object == NULL && Entry == &mData[Index] &&
						Entry->Next == NULL) || Entry->Object != NULL)
				if (Entry->Object == aObject)
					return true;
			}
			return false;
		}
		*/



		/*
		SEntry* GetEntry(T* aObject) const
		{
			Assert(aObject != NULL)
			unsigned int Index = ((TIntPtr)aObject) % mCapacity;
			for (SEntry* Entry = &mData[Index]; Entry != NULL; Entry = Entry->Next)
			{
				Assert((Entry->Object == NULL && Entry == &mData[Index] &&
						Entry->Next == NULL) || Entry->Object != NULL)
				if (Entry->Object == aObject)
					return Entry;
			}
			return NULL;
		}
		*/



		void AddObject(T* aObject)
		{
			Assert(aObject != NULL)
			unsigned int Index = ((TIntPtr)aObject) % mCapacity;
			if (mData[Index].Object == NULL)
			{
				Assert(mData[Index].Next == NULL)
				mData[Index] = SEntry(aObject);
			}
			else
			{
				SEntry *Entry;
				for (Entry = &mData[Index]; Entry->Next != NULL; Entry = Entry->Next)
				{
					Assert(Entry->Object != NULL)
					if (Entry->Object == aObject)
						throw CException("Object already in collection");
				}
				Entry->Next = new SEntry(aObject);
			}
		}
	};



	template <typename T> class CCollection
	{
	public:
		CCollection()
		{
			mCollection = new CCollectionIndex<T>();
			mIteratorEntry = NULL;
			#ifdef DEBUG
				mIterating = false;
			#endif
		}



		CCollection(const CCollection& aCollection)
		{
			mCollection = aCollection.mCollection->AddRef();
			mIteratorEntry = NULL;
			#ifdef DEBUG
				mIterating = false;
			#endif
		}



		~CCollection()
		{
			mCollection->Release();
		}



		void operator =(const CCollection& aCollection)
		{
			mCollection->Release();
			mCollection = aCollection.mCollection->AddRef();
			#ifdef DEBUG
				mIterating = false;
			#endif
		}



		int GetCount()
		{
			return mCollection->GetCount();
		}



		void Add(T* aObject)
		{
			MakeUnique();
			mCollection->Add(aObject);
			#ifdef DEBUG
				mIterating = false;
			#endif
		}



		void Remove(T* aObject)
		{
			MakeUnique();
			mCollection->Remove(aObject);
			#ifdef DEBUG
				mIterating = false;
			#endif
		}



		void First()
		{
			mCollection->IteratorFirst(mIteratorEntry, mIteratorIndex);
			mIteratorPrevEntry = NULL;
			#ifdef DEBUG
				mIterating = true;
				mIteratorValid = true;
			#endif
		}



		bool HasNext()
		{
			Assert(mIterating && mIteratorValid)
			#ifdef DEBUG
				if (mIteratorEntry == NULL)
					mIterating = false;
			#endif
			return mIteratorEntry != NULL;
		}



		void Next()
		{
			Assert(mIterating && (mIteratorEntry != NULL || !mIteratorValid))
			mCollection->IteratorNext(mIteratorEntry, mIteratorPrevEntry, mIteratorIndex);
			#ifdef DEBUG
				mIteratorValid = true;
			#endif
		}



		T* operator ->()
		{
			Assert(mIterating && mIteratorValid && mIteratorEntry != NULL)
			return mIteratorEntry->Object;
		}

		T* Current()
		{
			Assert(mIterating && mIteratorValid && mIteratorEntry != NULL)
			return mIteratorEntry->Object;
		}



		void Delete()
		{
			Assert(mIterating && mIteratorValid && mIteratorEntry != NULL)
			MakeUnique();
			mCollection->IteratorDelete(mIteratorEntry, mIteratorPrevEntry, mIteratorIndex);
			mIteratorEntry = NULL;
			#ifdef DEBUG
				mIteratorValid = false;
			#endif
		}

	private:
		CCollectionIndex<T> *mCollection;
		typename CCollectionIndex<T>::SEntry *mIteratorEntry;
		typename CCollectionIndex<T>::SEntry *mIteratorPrevEntry;
		unsigned int mIteratorIndex;
		#ifdef DEBUG
			bool mIterating;
			bool mIteratorValid;
		#endif



		void MakeUnique()
		{
			if (mCollection->IsShared())
			{
				CCollectionIndex<T> *Collection = new CCollectionIndex<T>(mCollection);
				mCollection->Release();
				mCollection = Collection;
			}
		}
	};
}