/*
CStringMap class
4 September 2007
Updated: 12 September 2009

Uses a hash table to store string keys.
*/

#include <vector>

namespace Obelisk
{
	template <typename V> struct SStringPair
	{
		SStringPair() { }
		SStringPair(CString aKey, V aValue)
				: Key(aKey), Value(aValue), Next(NULL)
		{
		}

		CString Key;
		V Value;
		SStringPair<V> *Next;
	};



	template <typename V> class CStringMapIndex
	{
	public:
		CStringMapIndex()
		{
			mReferenceCount = 1;
			mCapacity = 8;
			mRehashCapacity = mCapacity * 0.8;
			mData = new SStringPair<V>* [mCapacity];
			memset(mData, 0, mCapacity * sizeof(SStringPair<V>*));
			mNumItems = 0;
		}



		CStringMapIndex(const CStringMapIndex* qIndex)
		{
			mReferenceCount = 1;
		}



		~CStringMapIndex()
		{
			for (int i = 0; i < mCapacity; i++)
			{
				SStringPair<V> *e = mData[i];
				while (e != NULL)
				{
					SStringPair<V> *Next = e->Next;
					delete e;
					e = Next;
				}
			}
			delete [] mData;
		}



		CStringMapIndex* AddRef()
		{
			mReferenceCount++;
			return this;
		}



		void Release()
		{
			if (--mReferenceCount == 0)
				delete this;
		}



		void Add(CString aKey, V aValue)
		{
			if (ContainsKey(aKey))
				throw CException("Key already exists");
			Set(aKey, aValue);
		}

		void Set(CString aKey, V aValue)
		{
			unsigned int Index = Hash(aKey) % mCapacity;
			SStringPair<V> *LastEntry = NULL;
			for (SStringPair<V>* e = mData[Index]; e != NULL; e = e->Next)
			{
				LastEntry = e;
				if (e->Key == aKey)
				{
					e->Value = aValue;
					return;
				}
			}
			SStringPair<V> *Entry = new SStringPair<V>(aKey, aValue);
			if (LastEntry == NULL)
				mData[Index] = Entry;
			else
				LastEntry->Next = Entry;
			if (mNumItems == mRehashCapacity)
				Rehash();
			mNumItems++;
		}



		void Remove(CString aKey)
		{
			unsigned int Index = Hash(aKey) % mCapacity;
			SStringPair<V> *Last = NULL;
			for (SStringPair<V>* e = mData[Index]; e != NULL; e = e->Next)
			{
				if (e->Key == aKey)
				{
					if (Last == NULL)
						mData[Index] = e->Next;
					else
						Last->Next = e->Next;
					delete e;
					mNumItems--;
					// TODO: rehash if NumEntries is below a threshold
					return;
				}
				Last = e;
			}
			throw CException("Key does not exist");
		}



		void RemoveIfExists(CString aKey)
		{
			if (ContainsKey(aKey))
				Remove(aKey);
		}



		V& Lookup(CString aKey) const
		{
			unsigned int Index = Hash(aKey) % mCapacity;
			for (SStringPair<V>* e = mData[Index]; e != NULL; e = e->Next)
			{
				if (e->Key == aKey)
					return e->Value;
			}
			throw CException("Key does not exist");
		}



		bool ContainsKey(CString aKey) const
		{
			unsigned int Index = Hash(aKey) % mCapacity;
			for (SStringPair<V>* e = mData[Index]; e != NULL; e = e->Next)
			{
				if (e->Key == aKey)
					return true;
			}
			return false;
		}



		static bool StringMapIndexEquals(const CStringMapIndex<V>* a, const CStringMapIndex<V>* b)
		{
			if (a->mNumItems != b->mNumItems)
				return false;
			std::vector<CString> Keys = a->GetKeys();
			std::vector<V> Values = a->GetValues();
			for (int i = 0; i < Keys.size(); i++)
			{
				if (!b->ContainsKey(Keys[i]))
					return false;
				if (!(Values[i] == b->Lookup(Keys[i])))
					return false;
			}
			return true;
		}



		std::vector<CString> GetKeys() const
		{
			std::vector<CString> Keys;
			for (int i = 0; i < mCapacity; i++)
			{
				for (SStringPair<V> *e = mData[i]; e != NULL; e = e->Next)
					Keys.push_back(e->Key);
			}
			return Keys;
		}



		std::vector<V> GetValues() const
		{
			std::vector<V> Values;
			for (int i = 0; i < mCapacity; i++)
			{
				for (SStringPair<V> *e = mData[i]; e != NULL; e = e->Next)
					Values.push_back(e->Value);
			}
			return Values;
		}



		bool IsShared() const { return mReferenceCount != 1; }

	private:
		int mReferenceCount;
		int mCapacity;
		int mRehashCapacity;
		SStringPair<V> **mData;
		int mNumItems;



		unsigned int Hash(CString s) const
		{
			const TByte *Data = (const TByte*)s.GetData();
			int Size = s.GetSize() * sizeof(TCodeUnit);
			unsigned int Hash = 0;
			for (int i = 0; i < Size; i++)
			{
				Hash *= 7;
				Hash += Data[i] + 1;
			}
			return Hash;
		}



		/*
		Increases the capacity of the table and rebuilds it.
		*/
		void Rehash()
		{
			SStringPair<V> **OldData = mData;
			int OldCapacity = mCapacity;
			mCapacity += mCapacity;
			mData = new SStringPair<V>* [mCapacity];
			memset(mData, 0, mCapacity * sizeof(SStringPair<V>*));

			for (int i = 0; i < OldCapacity; i++)
			{
				SStringPair<V> *e = OldData[i];
				while (e != NULL)
				{
					SStringPair<V> *Next = e->Next;
					e->Next = NULL;
					int Index = Hash(e->Key) % mCapacity;
					SStringPair<V> *Last = mData[Index];
					if (Last == NULL)
						mData[Index] = e;
					else
					{
						while (Last->Next != NULL)
							Last = Last->Next;
						Last->Next = e;
					}

					e = Next;
				}
			}
			delete [] OldData;

			mRehashCapacity = mCapacity * 0.8;
		}
	};



	template <typename V> class CStringMap
	{
	public:
		CStringMap()
		{
			mIndex = new CStringMapIndex<V>();
		}
		~CStringMap()
		{
			mIndex->Release();
		}

		CStringMap(const CStringMap<V>& aMap)
		{
			mIndex = aMap.mIndex->AddRef();
		}

		CStringMap& operator =(const CStringMap<V>& aMap)
		{
			mIndex->Release();
			mIndex = aMap.mIndex->AddRef();
			return *this;
		}

		bool operator ==(const CStringMap<V>& aMap) const
		{
			return CStringMapIndex<V>::StringMapIndexEquals(mIndex, aMap.mIndex);
		}

		bool operator !=(const CStringMap<V>& aMap) const
		{
			return !CStringMapIndex<V>::StringMapIndexEquals(mIndex, aMap.mIndex);
		}

		void Add(CString aKey, V aValue)
		{
			if (mIndex->ContainsKey(aKey))
				throw CException("Key already exists");
			Set(aKey, aValue);
		}

		void Set(CString aKey, V aValue)
		{
			Uniquify();
			mIndex->Set(aKey, aValue);
		}

		void Remove(CString aKey)
		{
			Uniquify();
			mIndex->Remove(aKey);
		}

		void RemoveIfExists(CString aKey)
		{
			Uniquify();
			mIndex->RemoveIfExists(aKey);
		}

		V& operator [](CString aKey) const
		{
			return mIndex->Lookup(aKey);
		}

		bool ContainsKey(CString aKey) const
		{
			return mIndex->ContainsKey(aKey);
		}

		std::vector<CString> GetKeys() const
		{
			return mIndex->GetKeys();
		}

		std::vector<V> GetValues() const
		{
			return mIndex->GetValues();
		}

	private:
		CStringMapIndex<V> *mIndex;

		void Uniquify()
		{
			if (mIndex->IsShared())
			{
				mIndex->Release();
				mIndex = new CStringMapIndex<V>(mIndex);
			}
		}
	};
}